##
# File: BinaryCifReader.py
# Date: 15-May-2021  jdw
#
#  Reader methods and decoders for binary CIF.
#
#  Updates:
##

from collections import OrderedDict
import gzip
import io
import logging
import struct
from contextlib import closing

import msgpack
import requests
from mmcif.api.DataCategory import DataCategory
from mmcif.api.PdbxContainers import DataContainer

try:
    from urllib.parse import urlsplit
except Exception:
    from urlparse import urlsplit

logger = logging.getLogger(__name__)


class BinaryCifReader(object):
    """Reader methods for the binary CIF format."""

    def __init__(self, storeStringsAsBytes=False, defaultStringEncoding="utf-8"):
        """Create an instance of the binary CIF reader class.

        Args:
            storeStringsAsBytes (bool, optional): strings are stored as lists of bytes. Defaults to False.
            defaultStringEncoding (str, optional): default encoding for string data. Defaults to "utf-8".
        """
        self.__storeStringsAsBytes = storeStringsAsBytes
        self.__defaultStringEncoding = defaultStringEncoding

    def deserialize(self, locator):
        """Deserialize the input binary CIF file stored in the file/URL locator path.

        Args:
            locator (str): input file path or URL

        Returns:
            list: list DataContainer objects
        """
        cL = []
        try:
            if self.__isLocal(locator):
                with gzip.open(locator, mode="rb") if locator[-3:] == ".gz" else open(locator, "rb") as fh:
                    cL = self.__deserialize(fh, storeStringsAsBytes=self.__storeStringsAsBytes)
            else:
                if locator.endswith(".gz"):
                    customHeader = {"Accept-Encoding": "gzip"}
                    with closing(requests.get(locator, headers=customHeader)) as fh:
                        ufh = gzip.GzipFile(fileobj=io.BytesIO(fh.content))
                        cL = self.__deserialize(ufh, storeStringsAsBytes=self.__storeStringsAsBytes)
                else:
                    with closing(requests.get(locator)) as fh:
                        cL = self.__deserialize(io.BytesIO(fh.content), storeStringsAsBytes=self.__storeStringsAsBytes)

        except Exception as e:
            logger.exception("Failing with %s", str(e))
        return cL

    def __deserialize(self, fh, storeStringsAsBytes=False):
        cL = []
        try:
            dec = BinaryCifDecoders(storeStringsAsBytes=storeStringsAsBytes)
            bD = msgpack.unpack(fh)
            #
            logger.debug("bD.keys() %r", bD.keys())
            logger.debug("bD['dataBlocks'] %s", bD[self.__toBytes("dataBlocks")])
            #
            for dataBlock in bD[self.__toBytes("dataBlocks")]:
                header = self.__fromBytes(dataBlock[self.__toBytes("header")]) if self.__toBytes("header") in dataBlock else None
                logger.debug("header %r", header)
                logger.debug("dataBlock %r", dataBlock)
                #
                dc = DataContainer(header)
                categoryList = dataBlock[self.__toBytes("categories")] if self.__toBytes("categories") in dataBlock else []
                for category in categoryList:
                    catName = self.__fromBytes(category[self.__toBytes("name")])[1:]
                    colList = category[self.__toBytes("columns")]
                    logger.debug("catName %r columns %r", catName, colList)
                    colD = OrderedDict()
                    atNameList = []
                    for col in colList:
                        logger.debug("col.keys() %r", col.keys())
                        atName = self.__fromBytes(col[self.__toBytes("name")])
                        atData = col[self.__toBytes("data")]
                        logger.debug("atData encoding (%d) data (%d)", len(atData[self.__toBytes("encoding")]), len(atData[self.__toBytes("data")]))
                        atMask = col[self.__toBytes("mask")]
                        logger.debug("catName %r atName %r", catName, atName)
                        logger.debug(" >atData.data    %r", atData[self.__toBytes("data")])
                        logger.debug(" >atData.encoding (%d) %r", len(atData[self.__toBytes("encoding")]), atData[self.__toBytes("encoding")])
                        logger.debug(" >mask %r", atMask)
                        tVal = dec.decode(col[self.__toBytes("data")][self.__toBytes("data")], col[self.__toBytes("data")][self.__toBytes("encoding")])
                        if col[self.__toBytes("mask")]:
                            mVal = dec.decode(col[self.__toBytes("mask")][self.__toBytes("data")], col[self.__toBytes("mask")][self.__toBytes("encoding")])
                            tVal = ["?" if m == 2 else "." if m == 1 else d for d, m in zip(tVal, mVal)]
                        colD[atName] = tVal
                        atNameList.append(atName)
                    #
                    cObj = DataCategory(catName, attributeNameList=atNameList)
                    genL = [colGen for colGen in colD.values()]
                    for row in zip(*genL):
                        logger.debug("row %r", row)
                        cObj.append(row)
                    #
                    dc.append(cObj)
                cL.append(dc)
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        return cL

    def __isLocal(self, locator):
        """Returns true if input string can be interpreted as a local file path.

        Args:
            locator (str): url or path string

        Returns:
            bool: True if locator is a local path
        """
        try:
            locSp = urlsplit(locator)
            return locSp.scheme in ["", "file"]
        except Exception as e:
            logger.exception("For locator %r failing with %s", locator, str(e))
        return None

    def __toBytes(self, strVal):
        """Optional conversion of the input string to bytes according to the class setting (storeStringsAsBytes).

        Args:
            strVal (string): input string

        Returns:
            string or bytes: optionally converted string.
        """
        try:
            return strVal.encode(self.__defaultStringEncoding) if self.__storeStringsAsBytes else strVal
        except (UnicodeDecodeError, AttributeError):
            logger.exception("Bad type for %r", strVal)
            return strVal

    def __fromBytes(self, byteVal):
        """Optional conversion of the input value according to the class setting (storeStringsAsBytes).

        Args:
            byteVal (string): input byte object

        Returns:
            string: optionally converted input value
        """
        try:
            return byteVal.decode(self.__defaultStringEncoding) if self.__storeStringsAsBytes else byteVal
        except (UnicodeDecodeError, AttributeError):
            logger.exception("Bad type for %r", byteVal)
            return byteVal


class BinaryCifDecoders(object):
    """Column oriented Binary CIF decoders implementing
    StringArray, ByteArray, IntegerPacking, Delta, RunLength,
    FixedPoint, and  IntervalQuantization from the BinaryCIF
    specification described in:

    Sehnal D, Bittrich S, Velankar S, Koca J, Svobodova R, Burley SK, Rose AS.
    BinaryCIF and CIFTools-Lightweight, efficient and extensible macromolecular data management.
    PLoS Comput Biol. 2020 Oct 19;16(10):e1008247.
    doi: 10.1371/journal.pcbi.1008247. PMID: 33075050; PMCID: PMC7595629.

    and in the specification at https://github.com/molstar/BinaryCIF/blob/master/encoding.md

    and from the I/HM Python implementation at https://github.com/ihmwg/python-ihm[summary]

    """

    bCifCodeTypeD = {1: "integer_8", 2: "integer_16", 3: "integer_32", 4: "unsigned_integer_8", 5: "unsigned_integer_16", 6: "unsigned_integer_32", 32: "float_32", 33: "float_64"}
    """Binary CIF protocol internal data type codes to integer and float types
    """
    bCifTypeD = {
        "integer_8": {"struct_format_code": "b", "min": -0x7F - 1, "max": 0x7F},
        "integer_16": {"struct_format_code": "h", "min": -0x7FFF - 1, "max": 0x7FFF},
        "integer_32": {"struct_format_code": "i", "min": -0x7FFFFFFF - 1, "max": 0x7FFFFFFF},
        "unsigned_integer_8": {"struct_format_code": "B", "min": 0, "max": 0xFF},
        "unsigned_integer_16": {"struct_format_code": "H", "min": 0, "max": 0xFFFF},
        "unsigned_integer_32": {"struct_format_code": "I", "min": 0, "max": 0xFFFFFFFF},
        "float_32": {"struct_format_code": "f", "min": 1.175494351e-38, "max": 3.402823466e38},
        "float_64": {"struct_format_code": "d", "min": 2.2250738585072014e-308, "max": 1.7976931348623158e308},
    }
    """Binary CIF data type feature dictionary
    """

    def __init__(self, storeStringsAsBytes=False, defaultStringEncoding="utf-8", verbose=False):
        """Create an instance of the binary CIF encoder class.

        Args:
            storeStringsAsBytes (bool, optional): express keys and strings as byte types otherwise follow the default encoding. Defaults to False.
            defaultStringEncoding (str, optional): default encoding for string types. Defaults to "utf-8".
            verbose(bool, optional): provide tracking of type conversion issues. Defaults to False.
        """
        self.__storeStringsAsBytes = storeStringsAsBytes
        self.__defaultStringEncoding = defaultStringEncoding
        self.__verbose = verbose
        #
        self.__encodersMethodD = {
            "StringArray": self.stringArrayDecoder,
            "ByteArray": self.byteArrayDecoder,
            "IntegerPacking": self.integerPackingDecoder,
            "Delta": self.deltaDecoder,
            "RunLength": self.runLengthDecoder,
            "FixedPoint": self.fixedPointDecoder,
            "IntervalQuantization": self.intervalQuantizationDecoder,
        }

    def decode(self, colDataList, encodingDictList):
        """Return the decoded input data column using the input list of encodings

        Args:
            colDataList (list): column of data to be decoded
            encodingDictList (list): list of dictionary holding binary CIF encoding details
                                 elements described in the specification at
                                 https://github.com/molstar/BinaryCIF/blob/master/encoding.md

        Yields:
            list: decoded list of column data
        """
        for encoding in reversed(encodingDictList):
            encType = self.__fromBytes(encoding[self.__toBytes("kind")])
            colDataList = self.__encodersMethodD[encType](colDataList, encoding)
        return colDataList

    def stringArrayDecoder(self, colDataList, encodingDict):
        """Decode an array of strings stored as a concatenation of all unique
        strings, a list of offsets to construct the unique substrings, and indices into
        the offset array.

        Args:
            colDataList (list): column of data to be decoded
            encodingDict (dict): dictionary of binary CIF encoding details
                                 elements described in the specification at
                                 https://github.com/molstar/BinaryCIF/blob/master/encoding.md

        Yields:
            list: decoded list of string data
        """
        offsetList = list(self.decode(encodingDict[self.__toBytes("offsets")], encodingDict[self.__toBytes("offsetEncoding")]))
        lookupIndexIt = self.decode(colDataList, encodingDict[self.__toBytes("dataEncoding")])

        stringData = self.__fromBytes(encodingDict[self.__toBytes("stringData")])
        uniqueStringList = []
        for iBegin, iEnd in zip(offsetList, offsetList[1:]):
            uniqueStringList.append(stringData[iBegin:iEnd])
            logger.debug("iBegin %d iEnd %d %r ", iBegin, iEnd, stringData[iBegin:iEnd])

        for ii in lookupIndexIt:
            yield uniqueStringList[ii] if ii >= 0 else None

    def byteArrayDecoder(self, colDataList, encodingDict):
        """Decode input byte list into a list of integers/floats

        Args:
            colDataList (list): column of data to be decoded
            encodingDict (dict): dictionary of binary CIF encoding details
                                 elements described in the specification at
                                 https://github.com/molstar/BinaryCIF/blob/master/encoding.md

        Yields:
            list: decoded list of integer/float data
        """
        structKey = self.bCifCodeTypeD[encodingDict[self.__toBytes("type")]]
        structFormatCode = self.bCifTypeD[structKey]["struct_format_code"]
        count = len(colDataList) // struct.calcsize(structFormatCode)
        # struct.unpack() format string for little-endian  = < format_string code * counts
        return struct.unpack("<" + structFormatCode * count, colDataList)

    def __unsignedDecode(self, colDataList, encodingDict):
        upperLimit = self.bCifTypeD["unsigned_integer_8"]["max"] if encodingDict[self.__toBytes("byteCount")] == 1 else self.bCifTypeD["unsigned_integer_16"]["max"]
        ii = 0
        while ii < len(colDataList):
            value = 0
            tVal = colDataList[ii]
            while tVal == upperLimit:
                value += tVal
                ii += 1
                tVal = colDataList[ii]
            yield value + tVal
            ii += 1

    def __signedDecode(self, colDataList, encodingDict):
        upperLimit = self.bCifTypeD["integer_8"]["max"] if encodingDict[self.__toBytes("byteCount")] == 1 else self.bCifTypeD["integer_16"]["max"]
        lowerLimit = self.bCifTypeD["integer_8"]["min"] if encodingDict[self.__toBytes("byteCount")] == 1 else self.bCifTypeD["integer_16"]["min"]
        ii = 0
        while ii < len(colDataList):
            value = 0
            tVal = colDataList[ii]
            while tVal == upperLimit or tVal == lowerLimit:
                value += tVal
                ii += 1
                tVal = colDataList[ii]
            yield value + tVal
            ii += 1

    def integerPackingDecoder(self, colDataList, encodingDict):
        """Decode a (32-bit) integer list packed into 8- or 16-bit values.

        Args:
            colDataList (list): column of data to be decoded
            encodingDict (dict): dictionary of binary CIF encoding details
                                 elements described in the specification at
                                 https://github.com/molstar/BinaryCIF/blob/master/encoding.md

        Yields:
            list: decoded list of integer data

        """
        if encodingDict[self.__toBytes("isUnsigned")]:
            return self.__unsignedDecode(colDataList, encodingDict)
        else:
            return self.__signedDecode(colDataList, encodingDict)

    def deltaDecoder(self, colDataList, encodingDict):
        """Decode an integer list stored as a list of consecutive differences.

        Args:
            colDataList (list): column of data to be decoded
            encodingDict (dict): dictionary of binary CIF encoding details
                                 elements described in the specification at
                                 https://github.com/molstar/BinaryCIF/blob/master/encoding.md

        Yields:
            list: decoded list of integer data
        """
        val = encodingDict[self.__toBytes("origin")]
        for diff in colDataList:
            val += diff
            yield val

    def runLengthDecoder(self, colDataList, encodingDict):
        """Decode an integer list stored as pairs of (value, number of repeats).

        Args:
            colDataList (list): column of data to be decoded
            encodingDict (dict): dictionary of binary CIF encoding details
                                 elements described in the specification at
                                 https://github.com/molstar/BinaryCIF/blob/master/encoding.md

        Yields:
            list: decoded list of integer data
        """
        _ = encodingDict
        colDataList = list(colDataList)
        for ii in range(0, len(colDataList), 2):
            for _ in range(colDataList[ii + 1]):
                yield colDataList[ii]

    def fixedPointDecoder(self, colDataList, encodingDict):
        """Decode a floating point list stored as integers and a multiplicative factor.

        Args:
            colDataList (list): column of data to be decoded
            encodingDict (dict): dictionary of binary CIF encoding details
                                 elements described in the specification at
                                 https://github.com/molstar/BinaryCIF/blob/master/encoding.md

        Yields:
            list: decoded list of float data
        """
        factor = float(encodingDict[self.__toBytes("factor")])
        for val in colDataList:
            yield float(val) / factor

    def intervalQuantizationDecoder(self, colDataList, encodingDict):
        """Decode a list of 32-bit integers quantized within a given interval into a list of floats.

        Args:
            colDataList (list): column of data to be decoded
            encodingDict (dict): dictionary of binary CIF encoding details
                                 elements described in the specification at
                                 https://github.com/molstar/BinaryCIF/blob/master/encoding.md

        Yields:
            list: decoded list of float data

        """
        delta = float(encodingDict[self.__toBytes("max")] - encodingDict[self.__toBytes("min")]) / float(encodingDict[self.__toBytes("numSteps")] - 1.0)
        minVal = encodingDict[self.__toBytes("min")]
        for val in colDataList:
            yield minVal + delta * val

    def __toBytes(self, strVal):
        """Optional conversion of the input string to bytes according to the class setting (storeStringsAsBytes).

        Args:
            strVal (string): input string

        Returns:
            string or bytes: optionally converted string.
        """
        try:
            return strVal.encode(self.__defaultStringEncoding) if self.__storeStringsAsBytes else strVal
        except (UnicodeDecodeError, AttributeError):
            if self.__verbose:
                logger.exception("Bad type for %r", strVal)
            return strVal

    def __fromBytes(self, byteVal):
        """Optional conversion of the input value according to the class setting (storeStringsAsBytes).

        Args:
            byteVal (string): input byte object

        Returns:
            string: optionally converted input value
        """
        try:
            return byteVal.decode(self.__defaultStringEncoding) if self.__storeStringsAsBytes else byteVal
        except (UnicodeDecodeError, AttributeError):
            if self.__verbose:
                logger.exception("Bad type for %r", byteVal)
            return byteVal



##
# File:  PdbxExceptions.py
# Date:  2012-01-09  Jdw  Adapted from PdbxParser
#
##
__docformat__ = "google en"
__author__ = "John Westbrook"
__email__ = "john.westbrook@rcsb.org"
__license__ = "Apache 2.0"


class PdbxError(Exception):
    """Class for catch general errors"""


#    def __init__(self, msg):
#        super(PdbxError, self).__init__(msg)


class PdbxSyntaxError(Exception):
    """Class for catching syntax errors"""


#    def __init__(self, msg):
#        super(PdbxSyntaxError, self).__init__(msg)


#



##
# File: BuildDictionaryExec.py
# Date: 4-Apr-2021  jdw
#
#  CLI to materialize text dictionary files from dictionary components and include directives.
#
#  Updates:
##
__docformat__ = "google en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Apache 2.0"

import argparse
import logging
import os
import sys

from mmcif.api.DictionaryInclude import DictionaryInclude
from mmcif.io.IoAdapterPy import IoAdapterPy as IoAdapter

HERE = os.path.abspath(os.path.dirname(__file__))
TOPDIR = os.path.dirname(os.path.dirname(os.path.dirname(HERE)))

logging.basicConfig(level=logging.INFO, format="%(asctime)s [%(levelname)s]-%(module)s.%(funcName)s: %(message)s")
logger = logging.getLogger()


def main():
    parser = argparse.ArgumentParser()
    #
    #
    parser.add_argument("--op", default=None, required=True, help="Operation (build | get_version)")
    parser.add_argument("--input_dict_path", required=True, default=None, help="Path to dictionary generator file")
    parser.add_argument("--output_dict_path", default=None, help="Path to output dictionary text file")
    parser.add_argument("--cleanup", default=False, action="store_true", help="Remove include instruction categories after processing")
    args = parser.parse_args()
    #
    try:
        op = args.op.lower() if args.op else None
        inputDictPath = args.input_dict_path
        outputDictPath = args.output_dict_path
        cleanup = args.cleanup
    except Exception as e:
        logger.exception("Argument processing problem %s", str(e))
        parser.print_help(sys.stderr)
        exit(1)

    ##
    if op == "build" and inputDictPath and outputDictPath:
        dirPath = os.path.abspath(os.path.dirname(inputDictPath))
        logger.info("Starting dictionary path %s", dirPath)
        myIo = IoAdapter(raiseExceptions=True)
        containerList = myIo.readFile(inputFilePath=inputDictPath)
        logger.info("Starting dictionary container list length (%d)", len(containerList))
        dIncl = DictionaryInclude(dirPath=dirPath)
        inclL = dIncl.processIncludedContent(containerList, cleanup=cleanup)
        logger.info("Processed dictionary container length (%d)", len(inclL))
        ok = myIo.writeFile(outputFilePath=outputDictPath, containerList=inclL)
        logger.info("Operation completed with status %r", ok)
    elif op == "get_version" and inputDictPath:
        logger.setLevel(logging.ERROR)
        myIo = IoAdapter(raiseExceptions=True)
        containerList = myIo.readFile(inputFilePath=inputDictPath)
        dIncl = DictionaryInclude()
        inclL = dIncl.processIncludedContent(containerList)
        baseContainer = inclL[0]
        if baseContainer.exists("dictionary"):
            cObj = baseContainer.getObj("dictionary")
            version = cObj.getValueOrDefault("version", 0, None)
            print(version)


if __name__ == "__main__":
    main()



# File: setup.py
# Date: 17-Dec-2017
#
# Update:  17-Jan-2018 jdw - resolve python virtual env issues with Tox.
#           8-Aug-2018 jdw - add py3.7
#          14-May-2021 jdw - make requirements*.txt authoritative
#
import glob
import os
import platform
import re
import subprocess
import sys
from distutils.version import LooseVersion  # pylint: disable=no-name-in-module,import-error

from setuptools import Extension, find_packages, setup
from setuptools.command.build_ext import build_ext


class CMakeExtension(Extension):
    def __init__(self, name, sourcedir="", sources=None):
        sources = sources if sources else []
        Extension.__init__(self, name, sources=sources)
        self.sourcedir = os.path.abspath(sourcedir)


class CMakeBuild(build_ext):
    def run(self):
        try:
            out = subprocess.check_output(["cmake", "--version"])
        except OSError:
            raise RuntimeError("CMake >= 3.4 must be installed to build the following extensions: " + ", ".join(e.name for e in self.extensions))

        if platform.system() == "Windows":
            cmakeVersion = LooseVersion(re.search(r"version\s*([\d.]+)", out.decode()).group(1))
            if cmakeVersion < "3.4.0":
                raise RuntimeError("CMake >= 3.4.0 is required on Windows")

        for ext in self.extensions:
            self.build_extension(ext)

    def build_extension(self, ext):
        debug = True
        extdir = os.path.abspath(os.path.dirname(self.get_ext_fullpath(ext.name)))
        cmakeArgs = ["-DCMAKE_LIBRARY_OUTPUT_DIRECTORY=" + extdir, "-DPYTHON_EXECUTABLE=" + sys.executable]

        # We need to help cmake find the correct python for this virtual env -
        # ---
        libPath = None
        lsp = os.path.join(sys.exec_prefix, "lib", "libpython") + "*"
        lpL = glob.glob(lsp)
        if lpL:
            libPath = lpL[0]
        elif hasattr(sys, "base_exec_prefix"):
            lsp = os.path.join(sys.base_exec_prefix, "lib", "libpython") + "*"  # pylint: disable=no-member
            lpL = glob.glob(lsp)
            if lpL:
                libPath = lpL[0]
        if libPath:
            cmakeArgs += ["-DPYTHON_LIBRARY=" + libPath]
        else:
            print("------ WARNING could not locate python library")
        # ---
        inclPath = None
        isp = os.path.join(sys.exec_prefix, "include", "python") + "%s.%s" % (sys.version_info.major, sys.version_info.minor) + "*"
        ipL = glob.glob(isp)
        if ipL:
            inclPath = ipL[0]
        elif hasattr(sys, "base_exec_prefix"):
            isp = os.path.join(sys.base_exec_prefix, "include", "python") + "%s.%s" % (sys.version_info.major, sys.version_info.minor) + "*"  # pylint: disable=no-member
            ipL = glob.glob(isp)
            if ipL:
                inclPath = ipL[0]
        if inclPath:
            cmakeArgs += ["-DPYTHON_INCLUDE_DIR=" + inclPath]
        else:
            print("------ WARNING could not locate python include files")
        # ---
        cfg = "Debug" if self.debug else "Release"
        buildArgs = ["--config", cfg]

        if platform.system() == "Windows":
            cmakeArgs += ["-DCMAKE_LIBRARY_OUTPUT_DIRECTORY_{}={}".format(cfg.upper(), extdir)]
            if sys.maxsize > 2 ** 32:
                cmakeArgs += ["-A", "x64"]
            buildArgs += ["--", "/m"]
        else:
            cmakeArgs += ["-DCMAKE_BUILD_TYPE=" + cfg]
            buildArgs += ["--", "-j2"]

        env = os.environ.copy()
        env["CXXFLAGS"] = '{} -DVERSION_INFO=\\"{}\\"'.format(env.get("CXXFLAGS", ""), self.distribution.get_version())
        env["RUN_FROM_DISUTILS"] = "yes"
        #
        if not os.path.exists(self.build_temp):
            os.makedirs(self.build_temp)
        #
        if debug:
            print("------------- setup.py -----------------")
            print("Extension source path ", ext.sourcedir)
            print("CMAKE_ARGS ", cmakeArgs)
            print("self.build_temp ", self.build_temp)
            print("extdir", extdir)
            print("ext.name", ext.name)
            print("sys.executable", sys.executable)
            print("sys.exec_prefix", sys.exec_prefix)
            print("CXXFLAGS ", env["CXXFLAGS"])

        #
        subprocess.check_call(["cmake", ext.sourcedir] + cmakeArgs, cwd=self.build_temp, env=env)
        subprocess.check_call(["cmake", "--build", "."] + buildArgs, cwd=self.build_temp)


packages = []
thisPackage = "mmcif"

# Load packages from requirements*.txt
with open("requirements.txt", "r") as ifh:
    packagesRequired = [ln.strip() for ln in ifh.readlines()]

with open("requirements-test.txt", "r") as ifh:
    packagesTest = [ln.strip() for ln in ifh.readlines()]

with open("requirements-doc.txt", "r") as ifh:
    packagesDoc = [ln.strip() for ln in ifh.readlines()]

with open("README.md", "r") as ifh:
    longDescription = ifh.read()

with open("mmcif/__init__.py", "r") as ifh:
    version = re.search(r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]', ifh.read(), re.MULTILINE).group(1)

if not version:
    raise RuntimeError("Cannot find version information")

setup(
    name=thisPackage,
    version=version,
    description="mmCIF Core Access Library",
    long_description=longDescription,
    long_description_content_type="text/markdown",
    author="John Westbrook",
    author_email="john.westbrook@rcsb.org",
    url="http://mmcif.wwpdb.org",
    #
    license="Apache 2.0",
    classifiers=[
        # "Development Status :: 3 - Alpha",
        "Development Status :: 4 - Beta",
        # 'Development Status :: 5 - Production/Stable',
        "Intended Audience :: Developers",
        "Natural Language :: English",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.9",
    ],
    #
    install_requires=packagesRequired,
    tests_require=packagesTest,
    extras_require={"all": packagesRequired + packagesTest + packagesTest, "test": packagesTest, "docs": packagesTest},
    #
    packages=find_packages(exclude=["mmcif.tests", "tests.*"]),
    package_data={
        # If any package contains *.md or *.rst ...  files, include them:
        "": ["*.md", "*.rst", "*.txt", "*.h", "*.C", ".c", "*.cpp"]
    },
    test_suite="mmcif.tests",
    #
    ext_modules=[CMakeExtension("mmcif.core.mmciflib")],
    cmdclass=dict(build_ext=CMakeBuild),
    zip_safe=False,
    entry_points={
        "console_scripts": [
            "build_dict_cli=mmcif.io.BuildDictionaryExec:main",
        ]
    },
)



##
# File:    DictionaryInclude.py
# Author:  jdw
# Date:    11-August-2021
# Version: 0.001
#
# Updates:
##
"""
Implements PDBx/mmCIF dictionary composition using dictionary, category and item include instructions.
"""

import logging
import os
from collections import OrderedDict

try:
    from urllib.parse import urlsplit
except Exception:
    from urlparse import urlsplit

from mmcif.api.PdbxContainers import CifName
from mmcif.io.IoAdapterPy import IoAdapterPy

__docformat__ = "google en"
__author__ = "John Westbrook"
__email__ = "john.westbrook@rcsb.org"
__license__ = "Apache 2.0"


logger = logging.getLogger(__name__)


class DictionaryInclude(object):
    def __init__(self, **kwargs):
        #
        self.__itemNameRelatives = [
            "_item.name",
            "_item_examples.name",
            "_ndb_item_description.name",
            "_item_related.name",
            "_category_key.name",
            "_item_structure.name",
            "_item_methods.name",
            "_item_aliases.name",
            "_item_dependent.dependent_name",
            "_item_default.name",
            "_pdbx_item_examples.name",
            "_item_units.name",
            "_item_related.related_name",
            "_item_description.name",
            "_item_dependent.name",
            "_item_range.name",
            "_item_sub_category.name",
            "_pdbx_item_range.name",
            "_pdbx_item_linked.condition_child_name",
            "_ndb_item_examples.name",
            "_pdbx_item_value_condition.item_name",
            "_ndb_item_range.name",
            "_item_linked.child_name",
            "_pdbx_item_description.name",
            "_pdbx_item_context.item_name",
            "_pdbx_item_enumeration_details.name",
            "_pdbx_item_linked_group_list.child_name",
            "_pdbx_item_linked_group_list.parent_name",
            "_pdbx_item_value_condition_list.target_item_name",
            "_ndb_item_enumeration.name",
            "_pdbx_item_linked.child_name",
            "_pdbx_item_value_condition.dependent_item_name",
            "_pdbx_item_enumeration.name",
            "_item_linked.parent_name",
            "_pdbx_item_value_condition_list.dependent_item_name",
            "_item_type.name",
            "_item_type_conditions.name",
            "_pdbx_item_linked.parent_name",
            "_item_enumeration.name",
        ]
        self.__categoryIdRelatives = [
            "_category.id",
            "_category_key.id",
            "_pdbx_item_linked_group.category_id",
            "_pdbx_category_examples.id",
            "_item.category_id",
            "_pdbx_category_context.category_id",
            "_pdbx_item_linked_group_list.parent_category_id",
            "_category_group.category_id",
            "_pdbx_category_description.id",
            "_ndb_category_examples.id",
            "_category_examples.id",
            "_category_methods.category_id",
            "_ndb_category_description.id",
            "_pdbx_item_linked_group_list.child_category_id",
        ]
        #
        self.__cwd = os.getcwd()
        self.__dirPath = kwargs.get("dirPath", os.getcwd())
        logger.info("Local dictionary include path relative to %s", self.__dirPath)
        self.__dirStack = []
        self.__locatorIndexD = {}

    def processIncludedContent(self, containerList, cleanup=False):
        """Process any dictionary, category or item include instructions in any data containers in the
        input list of dictionary data and definition containers.

        Args:
            containerList (list): list of input PdbxContainer data or definition container objects
            cleanup (bool, optional): flag to remove generator category objects after parsing (default: False)

        Returns:
            (list): list of data and definition containers incorporating included content

        Examples:
            ```python
            pathDdlIncludeDictionary = "mmcif_ddl-generator.dic"
            myIo = IoAdapter(raiseExceptions=True)
            containerList = myIo.readFile(inputFilePath=pathDdlIncludeDictionary)
            logger.info("Starting container list length (%d)", len(containerList))
            dIncl = DictionaryInclude()
            inclL = dIncl.processIncludedContent(containerList)
            logger.info("Processed included container length (%d)", len(inclL))
            ```

        """
        includeD = self.__getIncludeInstructions(containerList, cleanup=cleanup)
        includeContentD = self.__fetchIncludedContent(includeD, cleanup=cleanup)
        return self.__addIncludedContent(containerList, includeContentD)

    def __addIncludedContent(self, containerList, includeContentD):
        """Incorporate included content described in the input dictionary of include instructions produced by
        internal method __getIncludeInstructions().

        Args:
            containerList (list): list of input PdbxContainer data or definition container objects
            includeContentD (dict): {"dictionaryIncludeDict": {dictionary_id: {...include details...}},
                                    "categoryIncludeDict": {dictionary_id: {category_id: {...include details... }}},
                                    "itemIncludeDict": {dictionary_id: {category_id: {itemName: {...include details...}}}}
                                    }

        Returns:
            (list): list of data and definition containers incorporating included content
        """
        # Index the current container list...
        cD = OrderedDict()
        datablockName = "unnamed_1"
        for container in containerList:
            if container.getType() == "data":
                datablockName = container.getName()
            # Handle potentially unconsolidated definitions --
            cD.setdefault(datablockName, OrderedDict()).setdefault(container.getName(), []).append(container)
        #
        #
        for datablockName in cD:
            if datablockName in includeContentD:
                if "replace" in includeContentD[datablockName]:
                    # Organize the replacements by name
                    replaceDefinitionD = OrderedDict()
                    replaceDataD = OrderedDict()
                    for container in includeContentD[datablockName]["replace"]:
                        if container.getType() == "definition":
                            replaceDefinitionD.setdefault(container.getName(), []).append(container)
                        else:
                            replaceDataD.setdefault(datablockName, []).append(container)
                    #
                    for rN, rL in replaceDefinitionD.items():
                        if rN in cD[datablockName]:
                            cD[datablockName][rN] = rL
                    # replace data sections in the base container
                    baseContainer = cD[datablockName][datablockName][0]
                    for rN, containerL in replaceDataD.items():
                        for container in containerL:
                            for nm in container.getObjNameList():
                                obj = container.getObj(nm)
                                baseContainer.replace(obj)
                    #
                if "extend" in includeContentD[datablockName]:
                    extendDataD = OrderedDict()
                    for container in includeContentD[datablockName]["extend"]:
                        if container.getType() == "definition":
                            cD.setdefault(datablockName, OrderedDict()).setdefault(container.getName(), []).append(container)
                        else:
                            extendDataD.setdefault(datablockName, []).append(container)
                    # extend data sections in the base container
                    baseContainer = cD[datablockName][datablockName][0]
                    for rN, containerL in extendDataD.items():
                        for container in containerL:
                            for nm in container.getObjNameList():
                                obj = container.getObj(nm)
                                if baseContainer.exists(nm):
                                    baseObj = baseContainer.getObj(nm)
                                    for ii in range(obj.getRowCount()):
                                        rowD = obj.getRowAttributeDict(ii)
                                        baseObj.append(rowD)
                                else:
                                    baseContainer.append(obj)
        #
        # Unwind the container index
        #
        fullL = []
        for datablockName in cD:
            for cL in cD[datablockName].values():
                fullL.extend(cL)
        #
        return fullL

    def __getIncludeInstructions(self, containerList, cleanup=False):
        """Extract include instructions from categories pdbx_include_dictionary,  pdbx_include_category, and pdbx_include_item.

        Args:
          containerList (list): list of input PdbxContainer data or definition container objects
          cleanup (optional, bool): flag to remove generator category objects after parsing (default: False)

        Returns:
          A dictionary containing the dictionary, category and and item level include details.
          For example,

            ```python
            {
            "dictionaryIncludeDict": {dictionary_id: {...include details...}},
            "categoryIncludeDict": {dictionary_id: {category_id: {...include details... }}},
            "itemIncludeDict": {dictionary_id: {category_id: {itemName: {...include details...}}}},
            }
            ```


        """
        includeD = OrderedDict()
        try:
            unNamed = 1
            for container in containerList:
                if container.getType() == "data":
                    dictionaryIncludeDict = OrderedDict()
                    categoryIncludeDict = OrderedDict()
                    itemIncludeDict = OrderedDict()
                    if container.getName():
                        datablockName = container.getName()
                    else:
                        datablockName = str(unNamed)
                        unNamed += 1
                    logger.debug("Adding data sections from container name %s  type  %s", datablockName, container.getType())
                    tl = container.getObj("pdbx_include_dictionary")
                    if tl is not None:
                        for row in tl.getRowList():
                            tD = OrderedDict()
                            for atName in ["dictionary_id", "dictionary_locator", "include_mode", "dictionary_namespace_prefix", "dictionary_namespace_prefix_replace"]:
                                tD[atName] = row[tl.getIndex(atName)] if tl.hasAttribute(atName) else None
                            dictionaryIncludeDict[tD["dictionary_id"]] = tD
                        #
                        tl = container.getObj("pdbx_include_category")
                        if tl is not None:
                            for row in tl.getRowList():
                                tD = OrderedDict()
                                for atName in ["dictionary_id", "category_id", "include_as_category_id", "include_mode"]:
                                    tD[atName] = row[tl.getIndex(atName)] if tl.hasAttribute(atName) else None
                                categoryIncludeDict.setdefault(tD["dictionary_id"], {}).setdefault(tD["category_id"], tD)
                        #
                        tl = container.getObj("pdbx_include_item")
                        if tl is not None:
                            for row in tl.getRowList():
                                tD = OrderedDict()
                                for atName in ["dictionary_id", "item_name", "include_as_item_name", "include_mode"]:
                                    tD[atName] = row[tl.getIndex(atName)] if tl.hasAttribute(atName) else None
                                categoryId = CifName.categoryPart(tD["item_name"])
                                itemIncludeDict.setdefault(tD["dictionary_id"], {}).setdefault(categoryId, {}).setdefault(tD["item_name"], tD)
                    if cleanup:
                        for catName in ["pdbx_include_dictionary", "pdbx_include_category", "pdbx_include_item"]:
                            if container.exists(catName):
                                container.remove(catName)
                    #
                    includeD[datablockName] = {
                        "dictionaryIncludeDict": dictionaryIncludeDict,
                        "categoryIncludeDict": categoryIncludeDict,
                        "itemIncludeDict": itemIncludeDict,
                    }
        except Exception as e:
            logger.exception("Include processing failing with %s", str(e))
        return includeD

    def __fetchIncludedContent(self, includeD, cleanup=False):
        """Fetch included content following the instructions encoded in the input data structure.

        Args:
            includeD (dict):  {"dictionaryIncludeDict": {dictionary_id: {...include details...}},
                               "categoryIncludeDict": {dictionary_id: {category_id: {...include details... }}},
                               "itemIncludeDict": {dictionary_id: {category_id: {itemName: {...include details...}}}},
                               }
            cleanup (optional, bool): flag to remove generator category objects after parsing (default: false)

        Returns:
            (dict): {datablockName: {"extend": [container,...], "replace": [container, ...]}, ... }

        """

        includeDataD = {}
        try:
            for datablockName, inclD in includeD.items():
                cL = []
                for dictName, iD in inclD["dictionaryIncludeDict"].items():
                    locator = iD["dictionary_locator"]
                    if locator in self.__locatorIndexD:
                        logger.info("Skipping redundant include for %r at %r", dictName, locator)
                        continue
                    self.__locatorIndexD[locator] = dictName
                    #
                    # --- Fetch the dictionary component -
                    #
                    updateStack = self.__isLocal(locator)
                    if updateStack:
                        if not self.__dirStack:
                            # top-level include case
                            self.__dirStack.append(os.path.abspath(self.__dirPath))

                        # embedded include case (push directory containing the locator)
                        if not os.path.isabs(locator):
                            # handle the relative path case -
                            locator = os.path.abspath(os.path.join(self.__dirStack[-1], locator))
                            logger.debug("modified local relative locator is %r", locator)
                        self.__dirStack.append(os.path.dirname(locator))
                        logger.debug("dirStack (%d) top %r", len(self.__dirStack), self.__dirStack[-1])
                    containerList = self.processIncludedContent(self.__fetchLocator(locator), cleanup=cleanup)
                    if updateStack:
                        # restore stack context
                        self.__dirStack.pop()
                    #
                    nsPrefix = iD["dictionary_namespace_prefix"]
                    nsPrefixReplace = iD["dictionary_namespace_prefix_replace"]
                    dictInclMode = iD["include_mode"]
                    dataIncludeMode = iD["data_include_mode"] if "data_include_mode" in iD else "extend"
                    catInclD = inclD["categoryIncludeDict"][dictName] if dictName in inclD["categoryIncludeDict"] else None
                    itemInclD = inclD["itemIncludeDict"][dictName] if dictName in inclD["itemIncludeDict"] else None
                    #
                    #  Do data sections first.
                    for container in containerList:
                        if container.getType() == "data":
                            logger.debug("Including data container %r with %r", container.getName(), container.getObjNameList())
                            cL.append((container, dataIncludeMode))
                    #
                    if catInclD or itemInclD:
                        # Process only explicitly included categories/items in the dictionary component
                        if catInclD:
                            for container in containerList:
                                if container.getType() == "data":
                                    continue
                                cName = container.getName()
                                catName = cName if container.isCategory() else CifName.categoryPart(cName)
                                #
                                if catName in catInclD:
                                    if container.isAttribute() and itemInclD and catName in itemInclD and cName in itemInclD[catName]:
                                        inclMode = itemInclD[catName][cName]["include_mode"] if itemInclD[catName][cName]["include_mode"] else dictInclMode
                                        cL.append((self.__renameItem(container, itemInclD[catName][cName]["include_as_item_name"]), inclMode))
                                    else:
                                        inclMode = catInclD[catName]["include_mode"] if catInclD[catName]["include_mode"] else dictInclMode
                                        cL.append((self.__renameCategory(container, catInclD[catName]["include_as_category_id"]), inclMode))
                        elif itemInclD:
                            # Process only explicitly included items exclusive of explicitly included categories in the dictionary component
                            for container in containerList:
                                if container.getType() == "data":
                                    continue
                                cName = container.getName()
                                catName = cName if container.isCategory() else CifName.categoryPart(cName)
                                #
                                if container.isAttribute() and catName in itemInclD and cName in itemInclD[catName]:
                                    inclMode = itemInclD[catName][cName]["include_mode"] if itemInclD[catName][cName]["include_mode"] else dictInclMode
                                    cL.append((self.__renameItem(container, itemInclD[catName][cName]["include_as_item_name"]), inclMode))
                    else:
                        # Process the full content of the dictionary component
                        for container in containerList:
                            if container.getType() == "data":
                                continue
                            cName = container.getName()
                            catName = cName if container.isCategory() else CifName.categoryPart(cName)
                            #
                            if container.isAttribute():
                                newName = self.__substituteItemPrefix(cName, nsPrefix, nsPrefixReplace)
                                cL.append((self.__renameItem(container, newName), dictInclMode))
                            else:
                                newName = self.__substituteCategoryPrefix(catName, nsPrefix, nsPrefixReplace)
                                cL.append((self.__renameCategory(container, newName), dictInclMode))
                #
                for container, inclMode in cL:
                    if inclMode == "replace":
                        includeDataD.setdefault(datablockName, {}).setdefault("replace", []).append(container)
                    elif inclMode == "extend":
                        logger.debug("%r extending with %r", datablockName, container.getName())
                        includeDataD.setdefault(datablockName, {}).setdefault("extend", []).append(container)
                #
            for nm in includeDataD:
                numReplace = len(includeDataD[nm]["replace"]) if "replace" in includeDataD[nm] else 0
                numExtend = len(includeDataD[nm]["extend"]) if "extend" in includeDataD[nm] else 0
                logger.debug("includeDataD %s replace (%d) extend (%d)", nm, numReplace, numExtend)
            #
        except Exception as e:
            logger.exception("Failing with %s", str(e))

        return includeDataD

    def __isLocal(self, locator):
        try:
            locSp = urlsplit(locator)
            return locSp.scheme in ["", "file"]
        except Exception as e:
            logger.error("Bad include file path (%r) : %s", locator, str(e))
        return False

    def __fetchLocator(self, locator, **kwargs):
        """"""
        try:
            # Locate non-absolute paths relative to the dictionary incude file
            if self.__isLocal(locator) and not os.path.isabs(locator):
                logger.info("locator is %r", locator)
                logger.info("dirStack (%d) top %r", len(self.__dirStack), self.__dirStack[-1])
                locator = os.path.abspath(os.path.relpath(locator, start=self.__dirStack[-1]))
            #
            containerList = []
            workPath = kwargs.get("workPath", None)
            enforceAscii = kwargs.get("enforceAscii", False)
            raiseExceptions = kwargs.get("raiseExceptions", True)
            useCharRefs = kwargs.get("useCharRefs", True)
            #
            myIo = IoAdapterPy(raiseExceptions=raiseExceptions, useCharRefs=useCharRefs)
            containerList = myIo.readFile(locator, enforceAscii=enforceAscii, outDirPath=workPath)
            logger.info("Fetched %r dictionary container length (%d)", locator, len(containerList) if containerList else 0)
            logger.debug("%r", [container.getName() for container in containerList])
        except Exception as e:
            logger.exception("Failing for %s with %s", locator, str(e))
        return containerList

    def __substituteCategoryPrefix(self, catName, curPrefix, newPrefix):
        return catName.replace(curPrefix, newPrefix, 1) if catName and catName.startswith(curPrefix) else catName

    def __substituteItemPrefix(self, itemName, curPrefix, newPrefix):
        atName = CifName.attributePart(itemName)
        atName = atName.replace(curPrefix, newPrefix, 1) if atName and atName.startswith(curPrefix) else atName
        catName = CifName.categoryPart(itemName)
        catName = catName.replace(curPrefix, newPrefix, 1) if atName and catName.startswith(curPrefix) else catName
        return CifName.itemName(catName, atName)

    def __renameItem(self, container, newItemName):
        if not container and not container.isAttribute() or not newItemName:
            return container
        #
        itemNameCur = container.getName()
        if itemNameCur == newItemName:
            return container
        #
        try:
            for item in self.__itemNameRelatives:
                catName = CifName.categoryPart(item)
                if container.exists(catName):
                    cObj = container.getObj(catName)
                    atName = CifName.attributePart(item)
                    if cObj.hasAttribute(atName):
                        for iRow in range(cObj.getRowCount()):
                            curVal = cObj.getValue(atName, iRow)
                            if curVal == itemNameCur:
                                cObj.setValue(newItemName, atName, iRow)
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        return container

    def __renameCategory(self, container, newCategoryName):
        if not container and not container.isCategory() or not newCategoryName:
            return container
        #
        catNameCur = container.getName()
        if catNameCur == newCategoryName:
            return container
        try:
            for item in self.__categoryIdRelatives:
                catName = CifName.categoryPart(item)
                if container.exists(catName):
                    cObj = container.getObj(catName)
                    atName = CifName.attributePart(item)
                    if cObj.hasAttribute(atName):
                        for iRow in range(cObj.getRowCount()):
                            testVal = cObj.getValue(atName, iRow)
                            if testVal == catNameCur:
                                cObj.setValue(newCategoryName, atName, iRow)
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        return container



##
# File:    IoAdapterBase.py
# Author:  J. Westbrook
# Date:    1-Aug-2017
# Version: 0.001 Initial version
#
# Updates:
#   13-Jan-2018 jdw move _getCategoryNameList() PdbxContainerBase class
#    6-Aug-2018 jdw add _setContainerProperties() to assign default container properties
#   10-Aug-2018 jdw add _uncompress() method (suppress xz as lzma is not support in py27)
#   11-Nov-2018 jdw add method to _chooseTemporaryPath()
##
"""
Base class presenting essential PDBx/mmCIF IO methods.

"""

__docformat__ = "google en"
__author__ = "John Westbrook"
__email__ = "john.westbrook@rcsb.org"
__license__ = "Apache 2.0"

import bz2
import datetime
import gzip
import io
import logging

# import lzma
import os
import shutil
import tempfile
import time
import zipfile

from mmcif.io.PdbxExceptions import PdbxError

logger = logging.getLogger(__name__)


class IoAdapterBase(object):
    """Base class presenting essential mmCIF I/O methods."""

    def __init__(self, *args, **kwargs):
        """General options controlling I/O method operations:

        Args:
            raiseExceptions (bool, optional): Flag to indicate that API errors should generate exceptions (True) or catch and log errors (default=False)
            maxInputLineLength (int, optional):  Default maximum input line length (default=4096)
            useCharRefs (bool, optional): Replace non-ascii characters with XML Character References (default=True)
            readEncodingErrors (str, optional): treatment of encoding errors at read time (default='ignore')
            timing (bool, optional):  log timing details for parsing and processing steps (default=False)
            verbose (bool,optional):  log verbose output from wrapped libraries

        """
        _ = args
        self._raiseExceptions = kwargs.get("raiseExceptions", False)
        self._maxInputLineLength = kwargs.get("maxInputLineLength", 4096)
        self._useCharRefs = kwargs.get("useCharRefs", True)
        self.__logFilePath = None
        self._debug = kwargs.get("debug", False)
        self._timing = kwargs.get("timing", False)
        self._verbose = kwargs.get("verbose", True)
        self._readEncodingErrors = kwargs.get("readEncodingErrors", "ignore")

    def readFile(self, *args, **kwargs):
        """Read file method. (abstract)

         Args:
            inputFilePath (string):  Input file path/uri
                   kw:  optional key-value arguments

        Returns:
            list of DataContainer Objects:  list of data or definition container objects
        """
        raise NotImplementedError("To be implemented in subclass")

    def writeFile(self, outputFilePath, containerList, **kwargs):
        """Write file method - (abstract)

        Args:
            outputFilePath (string):  output file path
            containerList (list of DataContainer objects):  list of data or definition containers objects for output

        Returns:
            bool: Completion status
        """
        raise NotImplementedError("To be implemented in subclass")

    def getReadDiags(self):
        """Return any diagnostics from the last read operation. (abstract)"""
        raise NotImplementedError("To be implemented in subclass")

    def _setContainerProperties(self, containerList, **kwargs):
        try:
            for container in containerList:
                for kw in kwargs:
                    container.setProp(kw, kwargs[kw])
        except Exception:
            return False

    def _getTimeStamp(self):
        utcnow = datetime.datetime.utcnow()
        ts = utcnow.strftime("%Y-%m-%d:%H:%M:%S")
        return ts

    def _getCategoryNameList(self, container, lastInOrder=None, selectOrder=None):
        """Return an ordered list of categories in the input container subject to
        input category name lists.

        Args:
           container (DataContainer object):  Input DataContainer object
           lastInOrder (list):  names of categories to be shifted to the end of the container.
           selectOrder (list):  preferred order of category names

        Returns:
           catNameList: list:  augmented category list or full list (default)
        """
        catNameList = []
        if lastInOrder:
            objNameList = container.getObjNameList()
            lastList = []
            for nm in objNameList:
                if nm in lastInOrder:
                    lastList.append(nm)
                    continue
                catNameList.append(nm)
            catNameList.extend(lastList)
        elif selectOrder:
            for nm in selectOrder:
                if container.exists(nm):
                    catNameList.append(nm)
        else:
            catNameList = container.getObjNameList()

        return catNameList

    def _setLogFilePath(self, filePath):
        """Set the log file path."""
        self.__logFilePath = filePath

    def _getLogFilePath(self):
        """Return current log file path."""
        return self.__logFilePath

    def _appendToLog(self, stList):
        """Append input string list to the current log file -"""
        if not self.__logFilePath:
            return False
        try:
            with open(self.__logFilePath, "a") as ofh:
                ofh.write("%s\n" % "\n".join(stList))
            return True
        except Exception as e:
            logger.debug("Failing with %s", str(e))
        return True

    def _logError(self, msg):
        """Convenience method to log error messages and optionally raise general exceptions (PdbxError)."""
        self._appendToLog([msg])
        if self._raiseExceptions:
            raise PdbxError(msg)
        else:
            logger.error(msg)

    def _readLogRecords(self):
        """Return the contents of the current log file as list of strings."""
        diagL = []
        try:
            with open(self.__logFilePath, "r") as ifh:
                for line in ifh:
                    diagL.append(line[:-1])
        except Exception as e:
            msg = "No logfile found %s (%s)" % (self.__logFilePath, str(e))
            diagL.append(msg)
            logger.debug(msg)

        return diagL

    def __getDiscriminator(self):
        """Internal method returning a string which can discriminate among default file names -"""
        return str(time.time() * 10000).replace(".", "_")

    def _chooseTemporaryPath(self, filePath, outDirPath=None):
        """Select a path for temporary files in the priority order
        outDirpath, directory containing the input filePath, current working directory,
        dynamically created temporary directory.

        These choices harmonize various legacy api behaviors.
        """
        if outDirPath:
            return outDirPath
        #
        for oPath in [os.path.dirname(filePath), ".", tempfile.gettempdir()]:
            if os.access(oPath, os.W_OK):
                return oPath

    def _getDefaultFileName(self, filePath, fileType="cif-parser", fileExt="log", outDirPath=None, verify=True):
        """Return default file path for the target input file subject to input attributes and the output path."""
        returnFilePath = None
        try:
            _, fn = os.path.split(filePath)
            bn, _ = os.path.splitext(fn)
            #
            ft = fileType if fileType else "temp"
            fex = fileExt if fileExt else "tmp"
            #
            sf = "_" + ft + "_P" + self.__getDiscriminator() + "." + fex
            #
            # pth = outDirPath if outDirPath else '.'
            pth = self._chooseTemporaryPath(filePath, outDirPath=outDirPath)
            #
            if verify:
                # test if pth is actually writable ?  Throw exception otherwise -
                #
                testfile = tempfile.TemporaryFile(dir=pth)
                testfile.close()
                #
            returnFilePath = os.path.join(pth, bn + sf)
        except Exception as e:
            if self._raiseExceptions:
                raise e
            else:
                logger.error("Failed creating default filename for %s type %s with %s", filePath, fileType, str(e))

        return returnFilePath

    def _fileExists(self, filePath):
        """Verify that input file path exists and is readable."""
        try:
            if not os.access(filePath, os.R_OK):
                msg = "Missing file %r" % filePath
                self._appendToLog([msg])
                logger.error(msg)
                #
                if self._raiseExceptions:
                    raise PdbxError(msg)
                return False
            else:
                logger.debug("Reading from file path %s", filePath)
                return True
        except Exception as e:
            msg = "File check error for %r with %s " % (filePath, str(e))
            self._appendToLog([msg])
            if self._raiseExceptions:
                raise PdbxError(msg)
            else:
                logger.error(msg)
        return False

    def _cleanupFile(self, test, filePath):
        """Remove the input file path subject to the input test condition."""
        try:
            if test:
                os.remove(filePath)
        except Exception:
            pass

    def _toAscii(self, inputFilePath, outputFilePath, chunkSize=5000, encodingErrors="ignore", readEncodingErrors="ignore"):
        """Encode input file to Ascii and write this to the target output file.   Handle encoding
        errors according to the input settting ('ignore', 'escape', 'xmlcharrefreplace').
        """
        try:
            startTime = time.time()
            chunk = []
            with io.open(inputFilePath, "r", encoding="utf-8", errors=readEncodingErrors) as ifh, io.open(outputFilePath, "w", encoding="ascii") as ofh:
                for line in ifh:
                    # chunk.append(line.encode('ascii', 'xmlcharrefreplace').decode('ascii'))
                    chunk.append(line.encode("ascii", encodingErrors).decode("ascii"))
                    if len(chunk) == chunkSize:
                        ofh.writelines(chunk)
                        chunk = []
                ofh.writelines(chunk)
            if self._timing:
                stepTime1 = time.time()
                logger.info("Timing text file %s encoded to as ascii in %.4f seconds", inputFilePath, stepTime1 - startTime)
            return True
        except Exception as e:
            msg = "Failing text ascii encoding for %s with %s" % (inputFilePath, str(e))
            self._appendToLog([msg])
            logger.error(msg)
            if self._raiseExceptions:
                raise PdbxError(msg)
        #
        return False

    def _uncompress(self, inputFilePath, outputDir):
        """Uncompress the input file if the path name has a recognized compression type file extension.file

        Return the file name of the uncompressed file (in outDir) or the original input file path.

        """
        try:
            startTime = time.time()
            _, fn = os.path.split(inputFilePath)
            bn, _ = os.path.splitext(fn)
            outputFilePath = os.path.join(outputDir, bn)
            if inputFilePath.endswith(".gz"):
                with gzip.open(inputFilePath, mode="rb") as inpF:
                    with io.open(outputFilePath, "wb") as outF:
                        shutil.copyfileobj(inpF, outF)
            elif inputFilePath.endswith(".bz2"):
                with bz2.open(inputFilePath, mode="rb") as inpF:
                    with io.open(outputFilePath, "wb") as outF:
                        shutil.copyfileobj(inpF, outF)
            # elif inputFilePath.endswith(".xz"):
            #    with lzma.open(inputFilePath, mode="rb") as inpF:
            #        with io.open(outputFilePath, "wb") as outF:
            #            shutil.copyfileobj(inpF, outF)
            elif inputFilePath.endswith(".zip"):
                with zipfile.ZipFile(inputFilePath, mode="rb") as inpF:
                    with io.open(outputFilePath, "wb") as outF:
                        shutil.copyfileobj(inpF, outF)
            else:
                outputFilePath = inputFilePath
            if self._timing:
                stepTime1 = time.time()
                logger.info("Timing text file %s uncompressed in %.4f seconds", inputFilePath, stepTime1 - startTime)
            #
        except Exception as e:
            msg = "Failing uncompress for file %s with %s" % (inputFilePath, str(e))
            self._appendToLog([msg])
            logger.exception(msg)
            if self._raiseExceptions:
                raise PdbxError(msg)

        logger.debug("Returning file path %r", outputFilePath)
        return outputFilePath



##
#
# File:     DataCategoryFormatted.py
# Original: 02-Feb-2009   jdw
#
# Update:
#   14-Nov-2012   jdw refactoring
#   17-Nov-2012   jdw self._rowList becomes data -
#   17-Dec-2012   jdw add quoting preference as constructor option.
#   01-Aug-2017   jdw migrate portions to public repo
#   12-Feb-2019   jdw tweak semi-colon quoted string formatting
##
"""

A subclass of DataCategory including additional formatting methods.

"""
from __future__ import absolute_import

import logging
import re

from six.moves import range
from six import string_types

from mmcif.api.DataCategory import DataCategory


__docformat__ = "google en"
__author__ = "John Westbrook"
__email__ = "john.westbrook@rcsb.org"
__license__ = "Apache 2.0"


logger = logging.getLogger(__name__)


class DataCategoryFormatted(DataCategory):
    """A subclass of DataCategory including additional formatting methods."""

    def __init__(self, dataCategoryObj, preferDoubleQuotes=True):
        self.__dcObj = dataCategoryObj
        super(DataCategoryFormatted, self).__init__(self.__dcObj.getName(), self.__dcObj.getAttributeList(), self.__dcObj.data)
        #
        self._currentRowIndex = 0
        self._currentAttribute = None
        #
        self.__avoidEmbeddedQuoting = False
        self.__preferDoubleQuotes = preferDoubleQuotes
        #
        # --------------------------------------------------------------------
        # any whitespace
        self.__wsRe = re.compile(r"\s")
        # self.__wsAndQuotesRe=re.compile(r"[\s'\"]")
        self.__wsAndQuotesRe = re.compile(r"[\s'\"#]")
        # any newline or carriage control
        self.__nlRe = re.compile(r"[\n\r]")
        #
        # single quote
        self.__sqRe = re.compile(r"[']")
        #
        self.__sqWsRe = re.compile(r"('\s)|(\s')")

        # double quote
        self.__dqRe = re.compile(r'["]')
        self.__dqWsRe = re.compile(r'("\s)|(\s")')
        #
        self.__intRe = re.compile(r"^[0-9]+$")
        self.__floatRe = re.compile(r"^-?(([0-9]+)[.]?|([0-9]*[.][0-9]+))([(][0-9]+[)])?([eE][+-]?[0-9]+)?$")
        #
        self.__dataTypeList = [
            "DT_NULL_VALUE",
            "DT_INTEGER",
            "DT_FLOAT",
            "DT_UNQUOTED_STRING",
            "DT_ITEM_NAME",
            "DT_DOUBLE_QUOTED_STRING",
            "DT_SINGLE_QUOTED_STRING",
            "DT_MULTI_LINE_STRING",
        ]
        self.__formatTypeList = [
            "FT_NULL_VALUE",
            "FT_NUMBER",
            "FT_NUMBER",
            "FT_UNQUOTED_STRING",
            "FT_QUOTED_STRING",
            "FT_QUOTED_STRING",
            "FT_QUOTED_STRING",
            "FT_MULTI_LINE_STRING",
        ]
        #
        # try:
        #    basestring
        # except NameError:
        #    basestring = str
        #
        # self.__string_types = basestring

    def __formatPdbx(self, inp):
        """Format input data following PDBx quoting rules -"""
        try:
            if inp is None:
                return ("?", "DT_NULL_VALUE")

            # pure numerical values are returned as unquoted strings
            # if (isinstance(inp, int) or self.__intRe.search(str(inp))):
            #
            try:
                if isinstance(inp, int) or self.__intRe.search(inp):
                    return ([str(inp)], "DT_INTEGER")
            except Exception:
                pass

            # if (isinstance(inp, float) or self.__floatRe.search(str(inp))):
            try:
                if isinstance(inp, float) or self.__floatRe.search(inp):
                    return ([str(inp)], "DT_FLOAT")
            except Exception:
                pass

            # null value handling -

            if inp == "." or inp == "?":
                return ([inp], "DT_NULL_VALUE")

            if inp == "":
                return (["."], "DT_NULL_VALUE")

            # Contains white space or quotes ?
            if not self.__wsAndQuotesRe.search(inp):
                # if inp.startswith("_"):
                if inp[0] in ["_"]:
                    return (self.__doubleQuotedList(inp), "DT_ITEM_NAME")
                elif inp[0] in ["[", "]", "$", "#", ";"]:
                    return (self.__doubleQuotedList(inp), "DT_DOUBLE_QUOTED_STRING")
                elif inp[:5].lower() in ["data_", "loop_", "save_"]:
                    return (self.__doubleQuotedList(inp), "DT_DOUBLE_QUOTED_STRING")
                else:
                    return ([str(inp)], "DT_UNQUOTED_STRING")
            else:
                if self.__nlRe.search(inp):
                    return (self.__semiColonQuotedList(inp), "DT_MULTI_LINE_STRING")
                else:
                    if self.__preferDoubleQuotes:
                        if self.__avoidEmbeddedQuoting:
                            # change priority to choose double quoting where possible.
                            if not self.__dqRe.search(inp) and not self.__sqWsRe.search(inp):
                                return (self.__doubleQuotedList(inp), "DT_DOUBLE_QUOTED_STRING")
                            elif not self.__sqRe.search(inp) and not self.__dqWsRe.search(inp):
                                return (self.__singleQuotedList(inp), "DT_SINGLE_QUOTED_STRING")
                            else:
                                return (self.__semiColonQuotedList(inp), "DT_MULTI_LINE_STRING")
                        else:
                            # change priority to choose double quoting where possible.
                            if not self.__dqRe.search(inp):
                                return (self.__doubleQuotedList(inp), "DT_DOUBLE_QUOTED_STRING")
                            elif not self.__sqRe.search(inp):
                                return (self.__singleQuotedList(inp), "DT_SINGLE_QUOTED_STRING")
                            else:
                                return (self.__semiColonQuotedList(inp), "DT_MULTI_LINE_STRING")
                    else:
                        if self.__avoidEmbeddedQuoting:
                            # change priority to choose double quoting where possible.
                            if not self.__sqRe.search(inp) and not self.__dqWsRe.search(inp):
                                return (self.__singleQuotedList(inp), "DT_SINGLE_QUOTED_STRING")
                            elif not self.__dqRe.search(inp) and not self.__sqWsRe.search(inp):
                                return (self.__doubleQuotedList(inp), "DT_DOUBLE_QUOTED_STRING")
                            else:
                                return (self.__semiColonQuotedList(inp), "DT_MULTI_LINE_STRING")
                        else:
                            # change priority to choose double quoting where possible.
                            if not self.__sqRe.search(inp):
                                return (self.__singleQuotedList(inp), "DT_SINGLE_QUOTED_STRING")
                            elif not self.__dqRe.search(inp):
                                return (self.__doubleQuotedList(inp), "DT_DOUBLE_QUOTED_STRING")
                            else:
                                return (self.__semiColonQuotedList(inp), "DT_MULTI_LINE_STRING")
        except Exception as e:
            logger.exception("Failing with %s on input %r %r", str(e), inp, type(inp))

        return ("?", "DT_NULL_VALUE")

    def __dataTypePdbx(self, inp):
        """Detect the PDBx data type -"""
        if inp is None:
            return "DT_NULL_VALUE"

        # pure numerical values are returned as unquoted strings
        # if isinstance(inp, int) or self.__intRe.search(str(inp)):
        if isinstance(inp, int) or (isinstance(inp, string_types) and self.__intRe.search(inp)):
            return "DT_INTEGER"

        # if isinstance(inp, float) or self.__floatRe.search(str(inp)):
        if isinstance(inp, float) or (isinstance(inp, string_types) and self.__floatRe.search(inp)):
            return "DT_FLOAT"

        # null value handling -

        if inp == "." or inp == "?":
            return "DT_NULL_VALUE"

        if inp == "":
            return "DT_NULL_VALUE"

        # Contains white space or quotes ?
        if not self.__wsAndQuotesRe.search(inp):
            if inp.startswith("_"):
                return "DT_ITEM_NAME"
            else:
                return "DT_UNQUOTED_STRING"
        else:
            if self.__nlRe.search(inp):
                return "DT_MULTI_LINE_STRING"
            else:
                if self.__avoidEmbeddedQuoting:
                    if not self.__sqRe.search(inp) and not self.__dqWsRe.search(inp):
                        return "DT_DOUBLE_QUOTED_STRING"
                    elif not self.__dqRe.search(inp) and not self.__sqWsRe.search(inp):
                        return "DT_SINGLE_QUOTED_STRING"
                    else:
                        return "DT_MULTI_LINE_STRING"
                else:
                    if not self.__sqRe.search(inp):
                        return "DT_DOUBLE_QUOTED_STRING"
                    elif not self.__dqRe.search(inp):
                        return "DT_SINGLE_QUOTED_STRING"
                    else:
                        return "DT_MULTI_LINE_STRING"

    def __singleQuotedList(self, inp):
        ll = []
        ll.append("'")
        ll.append(inp)
        ll.append("'")
        return ll

    def __doubleQuotedList(self, inp):
        ll = []
        ll.append('"')
        ll.append(inp)
        ll.append('"')
        return ll

    def __semiColonQuotedList(self, inp):
        ll = []
        ll.append("\n")
        if inp[-1] == "\n":
            ll.append(";")
            ll.append(inp)
            ll.append(";")
            ll.append("\n")
        else:
            ll.append(";")
            ll.append(inp)
            ll.append("\n")
            ll.append(";")
            ll.append("\n")

        return ll

    def getValueFormatted(self, attributeName=None, rowIndex=None):
        if attributeName is None:
            attribute = self._currentAttribute
        else:
            attribute = attributeName

        if rowIndex is None:
            rowI = self._currentRowIndex
        else:
            rowI = rowIndex

        if isinstance(attribute, self._stringTypes) and isinstance(rowI, int):
            try:
                fList, _ = self.__formatPdbx(self.data[rowI][self._attributeNameList.index(attribute)])
                return "".join(fList)
            except IndexError:
                logger.exception("attributeName %s rowI %r rowdata %r", attributeName, rowI, self.data[rowI])
                raise IndexError
            except Exception as e:
                logger.exception(" Failing with %s - AttributeName %s rowI %r rowdata %r", str(e), attributeName, rowI, self.data[rowI])
        else:
            logger.error(" Type error - AttributeName %r rowI %r rowdata %r", attributeName, rowI, self.data[rowI])
            logger.error(" Type error - string types %r", self._stringTypes)
            raise TypeError(attribute)

    def getValueFormattedByIndex(self, attributeIndex, rowIndex):
        try:
            fList, _ = self.__formatPdbx(self.data[rowIndex][attributeIndex])
            return "".join(fList)
        except IndexError:
            logger.exception("attributeIndex %r rowIndex %r rowdata %r", attributeIndex, rowIndex, self.data[rowIndex][attributeIndex])
            raise IndexError
        except Exception as e:
            logger.exception("Failing with %s  - attributeIndex %r rowIndex %r rowdata %r", str(e), attributeIndex, rowIndex, self.data[rowIndex][attributeIndex])
            raise e

    def getAttributeValueMaxLengthList(self, steps=1):
        mList = [0 for i in range(len(self._attributeNameList))]
        for row in self.data[::steps]:
            for indx in range(len(self._attributeNameList)):
                val = row[indx]
                if isinstance(val, self._stringTypes):
                    tLen = len(val)
                else:
                    tLen = len(str(val))

                mList[indx] = max(mList[indx], tLen)
        return mList

    def getFormatTypeList(self, steps=1):
        try:
            curFormatTypeList = []
            curDataTypeList = ["DT_NULL_VALUE" for i in range(len(self._attributeNameList))]
            for row in self.data[::steps]:
                for indx in range(len(self._attributeNameList)):
                    val = row[indx]
                    # print "index ",indx," val ",val
                    dType = self.__dataTypePdbx(val)
                    dIndx = self.__dataTypeList.index(dType)
                    # print "d type", dType, " d type index ",dIndx

                    cType = curDataTypeList[indx]
                    cIndx = self.__dataTypeList.index(cType)
                    cIndx = max(cIndx, dIndx)
                    curDataTypeList[indx] = self.__dataTypeList[cIndx]

            # Map the format types to the data types
            curFormatTypeList = []
            for dt in curDataTypeList:
                ii = self.__dataTypeList.index(dt)
                curFormatTypeList.append(self.__formatTypeList[ii])
        except Exception as e:
            logger.exception("Failing with %s", str(e))

        return curFormatTypeList, curDataTypeList
